"""Base classes for reference counting
"""

from typing import Any


class ReferenceCounter(object):
    """A base class for reference counters

    Attributes:
        reference_count (int): property getter, reference count
    """

    def __init__(self, prior_ref_count: int = 0):
        """Initialize this with an initial reference count.

        Args:
            prior_ref_count (int): the initial reference count. Default 0 if this object is sole responsible for the lifecycle of the resource.
        """
        self._ref_count: int = prior_ref_count + 1

    @property
    def reference_count(self) -> int:
        """Get the current reference count"""
        return self._ref_count

    def add_ref(self) -> None:
        """Manually increment the reference count.

        Users usually have no need to call this method. They may have to if they
        manage cases where one native handle wrapper uses another wrapper (and its underlying resource).
        """
        self._ref_count = self._ref_count + 1

    def decrement_ref(self) -> None:
        """Manually increment the reference count.

        Users usually have no need to call this method. They may have to if they
        manage cases where one native handle wrapper uses another wrapper (and its underlying resource).
        """
        self._ref_count = self._ref_count - 1


class NativeHandle(ReferenceCounter):
    """A base class for wrappers around otherwise "unmanaged" resources e.g. in a native library.

    Attributes:
        _handle (object): The handle (e.g. cffi pointer) to the native resource.
        _finalizing (bool): a flag telling whether this object is in its deletion phase. This has a use in some advanced cases with reverse callback, possibly not relevant in Python.
    """

    def __init__(self, handle: Any = None, prior_ref_count: int = 0):
        """initialize a reference counter for a resource handle, with an initial reference count.

        Args:
            handle (object): The handle (e.g. cffi pointer) to the native resource.
            prior_ref_count (int): the initial reference count. Default 0 if this NativeHandle is sole responsible for the lifecycle of the resource.
        """
        super(NativeHandle, self).__init__(prior_ref_count)
        # TODO checks
        self._finalizing: bool = False
        self._handle: Any = None
        if handle is None:
            return  # defer setting handle to the inheritor.
        self._set_handle(handle, prior_ref_count)

    def _set_handle(self, handle: Any, prior_ref_count: int = 0) -> None:
        """Sets a handle, after performing checks on its suitability as a handle for this object.

        Args:
            handle (object): The handle (e.g. cffi pointer) to the native resource.
            prior_ref_count (int): the initial reference count. Default 0 if this NativeHandle is sole responsible for the lifecycle of the resource.

        Raises:
            error message when a handle is not a valid object.
        """
        if not self._is_valid_handle(handle):
            raise RuntimeError("The specified handle argument is not a valid handle")
        self._handle = handle
        self._ref_count = prior_ref_count + 1

    def _is_valid_handle(self, handle: Any) -> bool:
        """Checks a handle on its suitability as a handle for this object.
        This method must be overriden by the inheritors.

        Args:
            handle (object): The handle (e.g. cffi pointer) to the native resource.
        """
        # See also https://stackoverflow.com/questions/4714136/how-to-implement-virtual-methods-in-python
        # May want to make this abstract using ABC - we'll see.
        raise NotImplementedError("method _is_valid_handle must be overriden by child classes")



"""Platform specific helpers to manage locating native dynamic libraries

This module hosts features similar to https://github.com/rdotnet/dynamic-interop-dll/blob/master/DynamicInterop/PlatformUtility.cs

"""

import os
import sys
from glob import glob

from ctypes.util import find_library as ctypes_find_library
from typing import List, Optional, Union


def library_short_filename(library_name: str, platform:Optional[str] = None) -> str:
    """Based on the library name, return the platform-specific expected library short file name

    Args:
        library_name (str): name of the library, for instance 'R', which results out of this 
            function  as 'libR.so' on Linux and 'R.dll' on Windows

    Raises:
        ValueError: invalid argument

    Returns:
        str: expected short file name for the library, for this platform
    """
    if platform is None:
        platform = sys.platform
    if library_name is None:
        raise ValueError("library_name cannot be None")
    else:
        if platform == "win32":
            return "{}.dll".format(library_name)
        elif platform == "linux":
            return "lib{}.so".format(library_name)
        elif platform == "darwin":
            return "lib{}.dylib".format(library_name)
        else:
            raise NotImplementedError(f"Platform '{platform}' is not (yet) supported")

def find_full_path(name: str, prefix:Optional[str]=None) -> Union[str, None]:
    """Find the full path of a library in under the python 
        installation directory, or as devised by ctypes.find_library

    Args:
        name (str): Library name, e.g. 'R' for the R programming language.

    Returns:
        Union[str, None]: First suitable library full file name.

    Examples:
        >>> from refcount.putils import *
        >>> find_full_path('gfortran')
        '/home/xxxyyy/anaconda3/envs/wqml/lib/libgfortran.so'
        >>> find_full_path('R')
        'libR.so'
    """    
    full_libpath = None
    if prefix is None:
        prefix = sys.prefix
    if name is None:
        return None
    else:
        lib_short_fname = library_short_filename(name)
        prefixed_lib_pat = os.path.join(prefix, "lib*", lib_short_fname)
        prefixed_libs = glob(prefixed_lib_pat)
        if prefixed_libs:
            full_libpath = prefixed_libs[0]
    if not full_libpath:
        full_libpath = ctypes_find_library(name)
    return full_libpath


# def find_full_paths(dll_short_name: str, directories: List[str] = None) -> List[str]:
#     """Find the full paths to library files, if they exist

#     Args:
#         dll_short_name (str): Short file name of the libary to search for, e.g. 'libgfortran.so'
#         directories (List[str], optional): directories under which to look for this file. Defaults to None.

#     Returns:
#         List[str]: zero or more matches, full paths to candidate files
#     """    
#     if directories is None:
#         directories = []
#     full_paths = [os.path.join(d, dll_short_name) for d in directories]
#     return [x for x in full_paths if os.path.exists(x)]


# def find_full_paths_env_var(
#     dll_short_name: str, env_var_name: str = "PATH"
# ) -> List[str]:
#     """Find the full paths to library files, if they exist

#     Args:
#         dll_short_name (str): Short file name of the libary to search for, e.g. 'libgfortran.so'
#         env_var_name (str, optional): [description]. Environment variable with paths to search under. Defaults to "PATH".

#     Returns:
#         List[str]: zero or more matches, full paths to candidate files
#     """
#     x = os.environ.get(env_var_name)
#     if x is not None:
#         search_paths = x.split(os.pathsep)
#     else:
#         search_paths = [""]
#     return find_full_paths(dll_short_name, search_paths)


def augment_path_env(
    added_paths: Union[str, List[str]], subfolder: str = None, to_env: str = "PATH", prepend:bool=False
) -> str:
    """Build a new list of directory paths, prepending prior to an existing env var with paths. 

    New paths are prepended only if they do already exist.

    Args:
        added_paths (Union[str,List[str]]): paths prepended
        subfolder (str, optional): Optional subfolder name to append to each in path prepended. Useful for 64/32 bits variations. Defaults to None.
        to_env (str, optional): Environment variable with existing Paths to start with. Defaults to 'PATH'.

    Returns:
        str: Content (set of paths), typically for a updating/setting an environment variable
    """
    path_sep = os.pathsep
    if isinstance(added_paths, str):
        added_paths = [added_paths]
    prior_path_env = os.environ.get(to_env)
    if prior_path_env is not None:
        prior_paths = prior_path_env.split(path_sep)
    else:
        prior_paths = []
    def _my_path_join(x, subfolder):  # avoid trailing path separator
        if subfolder is not None and subfolder != "":
            return os.path.join(x, subfolder)
        else:
            return x
    if subfolder is not None:
        added_paths = [_my_path_join(x, subfolder) for x in added_paths]
    added_paths = [x for x in added_paths if os.path.exists(x)]
    if prepend:
        new_paths = added_paths + prior_paths
    else:
        new_paths = prior_paths + added_paths
    # TODO: check for duplicate folders, perhaps.
    new_env_val = path_sep.join(new_paths)
    return new_env_val

# TODO: is that of any use still?? refactored out from uchronia and co. , but appears unused.
# def find_first_full_path(native_lib_file_name, readable_lib_name = "native library", env_var_name = ""):
#     if os.path.isabs(native_lib_file_name):
#         if (not os.path.exists(native_lib_file_name)):
#             raise FileNotFoundError("Could not find specified file {0} to load for {1}".format(native_lib_file_name, readable_lib_name))
#         return native_lib_file_name
#     if (native_lib_file_name is None or native_lib_file_name == ''):
#         raise FileNotFoundError("Invalid empty file name to load for {0}".format(readable_lib_name))
#     native_lib_file_name = _find_first_full_path(native_lib_file_name, env_var_name)
#     return native_lib_file_name

# def _find_first_full_path(short_file_name, env_var_name = ""):
#     if (none_or_empty(short_file_name)):
#         raise Exception("short_file_name")
#     lib_search_path_env_var = env_var_name
#     if (none_or_empty(lib_search_path_env_var)):
#         if(sys.platform == 'win32'):
#             lib_search_path_env_var = "PATH"
#         else:
#             lib_search_path_env_var =  "LD_LIBRARY_PATH"
#     candidates = find_full_path_env_var(short_file_name, lib_search_path_env_var)
#     if ((len(candidates) == 0) and (sys.platform == 'win32')):
#         if (os.path.exists(short_file_name)):
#             candidates = [short_file_name]
#     if (len(candidates) == 0):
#         raise FileNotFoundError("Could not find native library named '{0}' within the directories specified in the '{1}' environment variable".format(short_file_name, lib_search_path_env_var))
#     else:
#         return candidates[0]

# def find_full_path_env_var(dll_short_name, env_var_name="PATH"):
#     x = os.environ.get(env_var_name)
#     if x is not None:
#         search_paths = x.split(os.pathsep)
#     else:
#         search_pathsPathUpdater = [""]
#     return find_full_paths(dll_short_name, search_paths)

# def find_full_paths(dll_short_name, directories = []):
#     full_paths = [os.path.join(d, dll_short_name) for d in directories]
#     return [x for x in full_paths if os.path.exists(x)]

# def none_or_empty(x):
#     return (x is None or x == '')


# # The following is useful, but idiosyncratic. Consider and rethink.
def _win_architecture(platform:str=None):
    platform = sys.platform if platform is None else platform
    if platform == "win32":
        arch = os.environ["PROCESSOR_ARCHITECTURE"]
        return "64" if arch == 'AMD64' else '32'
    else:
        return ""


def build_new_path_env (from_env:str='LIBRARY_PATH', to_env:str='PATH', platform:str=None) -> str:
    """Propose an update to an existing environment variable, based on the path(s) specified in another environment variable. This function is effectively meant to be useful on Windows only.

    Args:
        from_env (str, optional): name of the source environment variable specifying the location(s) of custom libraries to load. Defaults to 'LIBRARY_PATH'.
        to_env (str, optional): environment variable to update, most likely the Windows PATH env var. Defaults to 'PATH'.

    Returns:
        str: the proposed updated content for the 'to_env' environment variable. 
    """
    platform = sys.platform if platform is None else platform
    path_sep = os.pathsep
    shared_lib_paths = os.environ.get(from_env)
    if shared_lib_paths is not None:
        # We could consider a call to a logger info here
        subfolder = _win_architecture()
        shared_lib_paths_vec = shared_lib_paths.split(path_sep)
        return augment_path_env(shared_lib_paths_vec, subfolder, to_env=to_env)
    else:
        print("WARNING: a function was called to look for environment variable '{0}' to update the environment variable '{1}', but was not found. This may be fine, but if the package fails to load because '{2}' is not found, this is a likely cause.".format(from_env, to_env, lib_short_fname))
        return ""

def update_path_windows (from_env:str='LIBRARY_PATH', to_env:str='PATH') -> None:
    """If called on Windows, append an environment variable, based on the path(s) specified in another environment variable. This function is effectively meant to be useful on Windows only.

    Args:
        from_env (str, optional): name of the source environment variable specifying the location(s) of custom libraries to load. Defaults to 'LIBRARY_PATH'.
        to_env (str, optional): environment variable to update, most likely the Windows PATH env var. Defaults to 'PATH'.

    Returns:
        None
    """
    if(sys.platform == 'win32'):
        os.environ[to_env] = build_new_path_env(from_env, to_env, sys.platform)








try:
    # New format
    # > Python3.7 ish
    from importlib.metadata import version # type: ignore
    __version__ = version("refcount")
except:
    # Older format
    # Older backport version
    try:
        from importlib_metadata import version # type: ignore
        __version__ = version("refcount")
    except:
        # Can't work out how to import version - set to Unknown
        __version__ = "Unknown - import error"



"""Implementation of reference counting classes for external resources accessed via interoperability software such as cffi
"""

from typing import Any, Callable, Dict, Optional, Union
from typing_extensions import TypeAlias 

from cffi import FFI
from refcount.base import NativeHandle

# This is a Hack. I cannot use FFI.CData in type hints.
# CffiData: TypeAlias = FFI().CData
CffiData: TypeAlias = Any
"""A dummy type to use in type hints for limited documentation purposes

FFI.CData is a type, but it seems it cannot be used in type hinting.
"""


class CffiNativeHandle(NativeHandle):
    """Reference counting wrapper class for CFFI pointers

    Attributes:
        _handle (object): The handle (e.g. cffi pointer) to the native resource.
        _type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
        _finalizing (bool): a flag telling whether this object is in its deletion phase. This has a use in some advanced cases with reverse callback, possibly not relevant in Python.
    """

    # Say you have a C API as follows:
    # * `void* create_some_object();`
    # * `dispose_of_some_object(void* obj);`
    # and accessing it using Python and CFFI. Users would use the `calllib` function:
    # aLibPointer = callib('mylib', 'create_some_object');
    # but at some point when done need to dispose of it:
    # callib('mylib', 'dispose_of_some_object', aLibPointer);
    # In practice in real systems one quickly ends up with cases
    # where it is unclear when to dispose of the object.
    # If you call the `dispose_of_some_object` function more
    # than once of too soon, you could easily crash the program.
    # CffiNativeHandle is designed to alleviate this headache by
    # using Matlab native reference counting of `handle` classes to reliably dispose of objects.

    # This class is originally inspired from a class with a similar purpose in C#. See https://github.com/rdotnet/dynamic-interop-dll

    def __init__(self, handle: "CffiData", type_id: str = None, prior_ref_count: int = 0):
        """Initialize a reference counter for a resource handle, with an initial reference count.

        Args:
            handle (object): The handle (e.g. cffi pointer) to the native resource.
            type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
            prior_ref_count (int): the initial reference count. Default 0 if this NativeHandle is sole responsible for the lifecycle of the resource.
        """
        super(CffiNativeHandle, self).__init__(handle, prior_ref_count)
        # TODO checks on handle
        self._type_id = type_id
        self._finalizing: bool = False
        self._handle: "CffiData" = None
        if handle is None:
            return  # defer setting the handle to the inheritor.
        self._set_handle(handle, prior_ref_count)

    def _is_valid_handle(self, h: "CffiData") -> bool:
        """Checks if the handle is a CFFI CData pointer, acceptable handle for this wrapper.

        Args:
            handle (object): The handle (e.g. cffi pointer) to the native resource.
        """
        return isinstance(h, FFI.CData)

    def __dispose_impl(self, decrement: bool) -> None:
        """An implementation of the dispose method in a 'Dispose' software pattern. Avoids cyclic method calls.

        Args:
            decrement (bool): indicating whether the reference count should be decreased. It should almost always be True except in very unusual use cases (argument is for possible future use).
        """
        if self.disposed:
            return
        if decrement:
            self._ref_count = self._ref_count - 1
        if self._ref_count <= 0:
            if self._release_handle():
                self._handle = None

    @property
    def disposed(self) -> bool:
        """Has the native object and memory already been disposed of.

        Returns:
            (bool): The underlying native handle has been disposed of from this wrapper
        """
        return self._handle is None

    @property
    def is_invalid(self) -> bool:
        """Is the underlying handle valid? In practice synonym with the disposed attribute.

        Returns:
            (bool): True if this handle is valid
        """
        return self._handle is None

    def _release_handle(self) -> bool:
        """Must of overriden. Method disposing of the object pointed to by the CFFI pointer (handle)

        Raises:
            NotImplementedError: thrown if this method is not overriden by inheritors

        Returns:
            bool: Overriding implementation should return True if the release of native resources was successful, False otherwise.
        """
        # See also https://stackoverflow.com/questions/4714136/how-to-implement-virtual-methods-in-python
        # May want to make this abstract using ABC - we'll see.
        raise NotImplementedError("method _release_handle must be overriden by child classes")

    def get_handle(self) -> Union["CffiData", None]:
        """Gets the underlying low-level CFFI handle this object wraps

        Returns:
            (Union[CffiData, None]): CFFI handle or None
        """
        return self._handle

    # TODO?
    # @property.getter
    # def ptr(self):
    #     """ Return the pointer to a cffi object """
    #     return self._handle

    @property
    def type_id(self) -> Optional[str]:
        """Return an optional type identifier for the underlying native type.

        This can be in practice useful to be more transparent about the underlying
        type obtained via a C API with opaque pointers (i.e. void*)

        Returns:
            str: optional type identifier
        """
        return self._type_id

    # @property
    # def obj(self):
    #     """ Return the native object pointed to (cffi object) """
    #     return self._handle

    def __str__(self) -> str:
        """ string representation """
        if self.type_id is None or self.type_id == "":
            return "CFFI pointer handle to a native pointer " + str(self._handle)
        return 'CFFI pointer handle to a native pointer of type id "' + self.type_id + '"'

    @property
    def ptr(self) -> 'CffiData':
        """ Return the pointer (cffi object) """
        return self._handle

    @property
    def obj(self) -> Any:
        """ Return the object pointed to (cffi object) """
        return self._handle[0]


    def __repr__(self) -> str:
        """ string representation """
        return str(self)

    def __del__(self) -> None:
        """ destructor, triggering the release of the underlying handled resource if the reference count is 0 """
        if self._handle is not None:
            # if not self._release_native is None:
            # Protect against accessing properties
            # of partially constructed objects (May not be an issue in Python?)
            self._finalizing = True
            self.release()

    def dispose(self) -> None:
        """Disposing of the object pointed to by the CFFI pointer (handle) if the reference counts allows it."""
        self.__dispose_impl(True)

    def release(self) -> None:
        """Manually decrements the reference counter. Triggers disposal if reference count is down to zero."""
        self.__dispose_impl(True)


class DeletableCffiNativeHandle(CffiNativeHandle):
    """Reference counting wrapper class for CFFI pointers

    Attributes:
        _handle (object): The handle (e.g. cffi pointer) to the native resource.
        _type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
        _finalizing (bool): a flag telling whether this object is in its deletion phase. This has a use in some advanced cases with reverse callback, possibly not relevant in Python.
        _release_native (Callable[[CffiData],None]): function to call on deleting this wrapper. The function should have one argument accepting the object _handle.
    """

    def __init__(
        self,
        handle: "CffiData",
        release_native: Optional[Callable[["CffiData"], None]],
        type_id: str = None,
        prior_ref_count: int = 0,
    ):
        """New reference counter for a CFFI resource handle.

        Args:
            handle (CffiData): The handle (expected cffi pointer) to the native resource.
            release_native (Callable[[CffiData],None]): function to call on deleting this wrapper. The function should have one argument accepting the object handle.
            type_id (str, optional): [description]. An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation. Defaults to None.
            prior_ref_count (int, optional): [description]. The initial reference count. Defaults to 0 if this NativeHandle is sole responsible for the lifecycle of the resource.
        """
        super(DeletableCffiNativeHandle, self).__init__(
            handle, type_id, prior_ref_count
        )
        self._release_native = release_native
        self._set_handle(handle, prior_ref_count)

    def _release_handle(self) -> bool:
        """Release the handle, dispose of the native resource.

        Returns:
            bool: Return True if the release of native resources handle was successful, False otherwise.
        """
        if self._handle is None:
            return False
        if self._release_native is None:
            return False
        if self._release_native is not None:
            self._release_native(
                self._handle
            )  # TODO are trapped exceptions acceptable here?
            return True


class OwningCffiNativeHandle(CffiNativeHandle):
    """Reference counting wrapper class for CFFI pointers that own and already manage the native memory

    Attributes:
        _handle (object): The handle (e.g. cffi pointer) to the native resource.
        _type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
        _finalizing (bool): a flag telling whether this object is in its deletion phase. This has a use in some advanced cases with reverse callback, possibly not relevant in Python.
    """

    # """ a global function that can be called to release an external pointer """
    # release_native = None

    def __init__(
        self,
        handle: "CffiData",
        type_id: str = None,
        prior_ref_count: int = 0,
    ):
        """Reference counting wrapper class for CFFI pointers that own and already manage the native memory

        Args:
            handle (CffiData): The handle (expected cffi pointer) to the native resource.
            type_id (str, optional): [description]. An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation. Defaults to None.
            prior_ref_count (int, optional): [description]. The initial reference count. Defaults to 0 if this NativeHandle is sole responsible for the lifecycle of the resource.
        """
        super(OwningCffiNativeHandle, self).__init__(
            handle, type_id, prior_ref_count
        )
        self._set_handle(handle, prior_ref_count)

    def _release_handle(self) -> bool:
        """Does nothing, as the wrapped cffi pointer is already owning and managing the memory

        Returns:
            bool: Always returns True.
        """
        return True


def wrap_cffi_native_handle(
    obj: Union["CffiData", Any],
    type_id: str = "",
    release_native: Callable[["CffiData"], None] = None,
) -> Union[DeletableCffiNativeHandle, Any]:
    """Create a reference counting wrapper around an object if this object is a CFFI pointer

    Args:
        obj (Union[CffiData,Any]): An object, which will be wrapped if this is a CFFI pointer, i.e. an instance of `CffiData`
        release_native (Callable[[CffiData],None]): function to call on deleting this wrapper. The function should have one argument accepting the object handle.
        type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
    """
    if isinstance(obj, FFI.CData):
        return DeletableCffiNativeHandle(
            obj, release_native=release_native, type_id=type_id
        )
    else:
        return obj


def is_cffi_native_handle(x: Any, type_id: str = "") -> bool:
    """Checks whether an object is a ref counting wrapper around a CFFI pointer

    Args:
        x (object): object to test, presumed to be an instance of `CffiNativeHandle`
        type_id (Optional[str]): Optional identifier for the type of underlying resource being wrapped.
    """
    if x is None:
        return False
    if not isinstance(x, CffiNativeHandle):
        return False
    if type_id is None or type_id == "":
        return True
    return x.type_id == type_id


def unwrap_cffi_native_handle(
    obj_wrapper: Any, stringent: bool = False
) -> Union["CffiData", Any, None]:
    """Unwrap a reference counting wrapper and returns its CFFI pointer if it is found (wrapped or 'raw')

    Args:
        obj_wrapper (Any): An object, which will be unwrapped if this is a CFFI pointer, i.e. an instance of `CffiData`
        stringent (bool, optional): [description]. if True an error is raised if obj_wrapper is neither None, a CffiNativeHandle nor an CffiData. Defaults to False.

    Raises:
        Exception: A CFFI pointer could not be found in the object.

    Returns:
        Union[CffiData,Any,None]: A CFFI pointer if it was found. Returns None or unchanged if not found, and stringent is equal to False. Exception otherwise.
    """
    # 2016-01-28 allowing null pointers, to unlock behavior of EstimateERRISParameters.
    # Reassess approach, even if other C API function will still catch the issue of null ptrs.
    if obj_wrapper is None:
        return None
    if isinstance(obj_wrapper, CffiNativeHandle):
        return obj_wrapper.get_handle()
    if isinstance(obj_wrapper, FFI.CData):
        return obj_wrapper
    else:
        if stringent:
            raise TypeError(
                "Argument is neither a CffiNativeHandle nor a CFFI external pointer"
            )
        else:
            return obj_wrapper


def cffi_arg_error_external_obj_type(x: Any, expected_type_id: str) -> str:
    """Build an error message that an unexpected object is in lieu of an expected refcount external ref object.

    Args:
        x (object): object passed as an argument to a function but with an unexpected type or type id.
        expected_type_id (Optional[str]): Expected identifier for the type of underlying resource being wrapped.

    Returns (str): the error message
    """
    if x is None:
        return "Expected a 'CffiNativeHandle' but instead got 'None'"
    if not is_cffi_native_handle(x):
        return (
            "Expected a 'CffiNativeHandle' but instead got object of type '{0}'".format(
                str(type(x))
            )
        )
    else:
        return "Expected a 'CffiNativeHandle' with underlying type id '{0}' but instead got one with type id '{1}'".format(
            expected_type_id, x.type_id
        )


# Maybe, pending use cases:
# def checked_unwrap_cffi_native_handle (obj_wrapper, stringent=False):
#     if not is_cffi_native_handle (obj_wrapper, expected_type_id):
#         raise Exception(cffi_arg_error_external_obj_type(obj_wrapper, expected_type_id)
#     else:
#         return unwrap_cffi_native_handle (obj_wrapper, stringent=True)

class GenericWrapper:
    """A pass-through wrapper for python objects that are ready for C interop. "bytes" can be passed as C 'char*'

    This is mostly a facility to generate glue code more easily 
    """    

    def __init__(self, handle: Any):
        self._handle = handle

    @property
    def ptr(self) -> Any:
        return self._handle

def wrap_as_pointer_handle(
    obj_wrapper: Any, stringent: bool = False
) -> Union[CffiNativeHandle, OwningCffiNativeHandle, GenericWrapper, None]:
    """Wrap an object, if need be, so that its C API pointer appears accessible via a 'ptr' property

    Args:
        obj_wrapper (Any): Object to wrap, if necessary
        stringent (bool, optional): Throws an exception if the input type is unhandled. Defaults to False.

    Raises:
        TypeError: neither a CffiNativeHandle nor a CFFI external pointer, nor bytes

    Returns:
        Union[CffiNativeHandle, OwningCffiNativeHandle, GenericWrapper, None]: wrapped object or None
    """    
    # 2016-01-28 allowing null pointers, to unlock behavior of EstimateERRISParameters.
    # Reassess approach, even if other C API function will still catch the issue of null ptrs.
    if obj_wrapper is None:
        return None
        # return GenericWrapper(FFI.NULL)  # Ended with kernel crashes and API call return, but unclear why
    elif isinstance(obj_wrapper, CffiNativeHandle):
        return obj_wrapper
    elif isinstance(obj_wrapper, FFI.CData):
        return OwningCffiNativeHandle(obj_wrapper)
    elif isinstance(obj_wrapper, bytes):
        return GenericWrapper(obj_wrapper)
    else:
        if stringent:
            raise TypeError(
                "Argument is neither a CffiNativeHandle nor a CFFI external pointer, nor bytes"
            )
        else:
            return obj_wrapper


def type_error_cffi(x:Union[CffiNativeHandle, Any], expected_type:str) -> str:
    """DEPRECATED Build an error message for situations where a cffi pointer handler is not that, or not of the expected type

    Args:
        x (Union[CffiNativeHandle, Any]): actual object that is not of the expected type or underlying type for the external pointer.
        expected_type (str): underlying type expected for the CFFI pointer handler

    Returns:
        str: error message that the caller can use to report the issue
    """
    return cffi_arg_error_external_obj_type(x, expected_type)

class CffiWrapperFactory:
    """A class that creates custom python wrappers based on the type identifier of the external pointer being wrapped
    """
    def __init__(self, api_type_wrapper:Dict[str,Any], strict_wrapping:bool=False) -> None:
        """_summary_

        Args:
            api_type_wrapper (Dict[str,Any]): dictionary, mapping from type identifiers to callables, class constructors
            strict_wrapping (bool, optional): If true, type identifiers passed at wrapper creation time `create_wrapper` 
                must be known or exceptions are raised. If False, it falls back on creating generic wrappers. Defaults to False.
        """        
        self._strict_wrapping = strict_wrapping
        self._api_type_wrapper = api_type_wrapper

    def create_wrapper(self, obj: Any, type_id: str, release_native: Optional[Callable[["CffiData"], None]]) -> 'CffiNativeHandle':
        """_summary_

        Args:
            obj (Union[CffiData,Any]): An object, which will be wrapped if this is a CFFI pointer, i.e. an instance of `CffiData`
            type_id (Optional[str]): An optional identifier for the type of underlying resource. This can be used to usefully maintain type information about the pointer/handle across an otherwise opaque C API. See package documentation.
            release_native (Callable[[CffiData],None]): function to call on deleting this wrapper. The function should have one argument accepting the object handle.

        Raises:
            ValueError: Missing type_id
            ValueError: If this object is in strict mode, and `type_id` is not known in the mapping
            NotImplementedError: `type_id` is known, but mapping to None (wrapper not yet implemented)
            TypeError: The function to create the wrapper does not accept any argument. 

        Returns:
            CffiNativeHandle: cffi wrapper
        """        
        from inspect import signature
        if type_id is None:
            raise ValueError("Type ID provided cannot be None")
        if type_id not in self._api_type_wrapper.keys():
            if self._strict_wrapping:
                raise ValueError("Type ID {} is unknown".format(type_id))
            else:
                return wrap_cffi_native_handle(obj, type_id, release_native)
        wrapper_type = self._api_type_wrapper[type_id]
        if wrapper_type is None:
            if self._strict_wrapping:
                raise NotImplementedError(
                    "Python object wrapper for foreign type ID {} is not yet implemented".format(
                        wrapper_type
                    )
                )
            else:
                return wrap_cffi_native_handle(obj, type_id, release_native)
        else:
            s = signature(wrapper_type)
            n = len(s.parameters)
            if n == 0:
                raise TypeError("Wrapper constructor must have at least one argument")
            elif n == 1:
                return wrapper_type(obj)
            elif n == 2:
                return wrapper_type(obj, release_native)
            else:
                return wrapper_type(obj, release_native, type_id)


WrapperCreationFunction = Callable[[Any, str, Callable], DeletableCffiNativeHandle]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['refcount']

package_data = \
{'': ['*']}

install_requires = \
['cffi>=1.11.5,<2.0.0', 'typing-extensions>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'refcount',
    'version': '1.1.1',
    'description': 'Python classes for reference counting',
    'long_description': '# refcount - Python classes for reference counting\n\n[![license](https://img.shields.io/badge/license-MIT-blue.svg)](https://github.com/csiro-hydroinformatics/pyrefcount/blob/master/LICENSE.txt) ![status](https://img.shields.io/badge/status-stable-green.svg) [![Documentation Status](https://readthedocs.org/projects/pyrefcount/badge/?version=latest)](https://pyrefcount.readthedocs.io/en/latest/?badge=latest) [![codecov](https://codecov.io/gh/csiro-hydroinformatics/pyrefcount/branch/master/graph/badge.svg?token=ZBBFq3ncAq)](https://codecov.io/gh/csiro-hydroinformatics/pyrefcount) master: [![Python package](https://github.com/csiro-hydroinformatics/pyrefcount/actions/workflows/build-matrix.yml/badge.svg?branch=master)](https://github.com/csiro-hydroinformatics/pyrefcount/actions/workflows/build-matrix.yml) testing: [![Python package](https://github.com/csiro-hydroinformatics/pyrefcount/actions/workflows/build-matrix.yml/badge.svg?branch=testing)](https://github.com/csiro-hydroinformatics/pyrefcount/actions/workflows/build-matrix.yml)\n\n\n![Reference counted native handles](./docs/img/refcount-principles.png "Reference counted native handles")\n\nThis package is primarily for managing resources in native libraries, written for instance in C++, from Python. While it boils down to "simply" maintaining a set of counters, **it is deceptively complicated to do so properly** and not end up with memory leaks or crashes. This package offers structured options for reliably managing external native resources. Surprisingly I could not locate an existing package doing just what I needed. Other use cases requiring reference counting, aside from native library resources, may benefit from reusing and extending classes in `refcount`.\n\n`refcount` ( >=0.7) includes classes using [cffi](https://cffi.readthedocs.io/). Other low-level interoperability mechanisms may well be added in the future.\n\n## License\n\nMIT (see [License.txt](https://github.com/csiro-hydroinformatics/pyrefcount/blob/master/LICENSE.txt))\n\n## Documentation\n\nHosted at [refcount via readthedocs.io](https://pyrefcount.readthedocs.io/en/latest/?badge=latest)\n\n## Source code\n\nThe code repository is on [GitHub](https://github.com/csiro-hydroinformatics/pyrefcount).\n\n## Installation\n\n### conda-forge\n\nUsing `conda` or `mamba`:\n\n```sh\nmamba install -c conda-forge refcount\n```\n\n### pypi\n\n```sh\npip install refcount\n```\n\n### From source (development)\n\n```sh\npip install -r requirements.txt\npip install -e .\n```\n\n## Sample use\n\nThe following example is based on one of the unit tests.\n\nSay we have a C++ library with objects and a C API:\n\n```C++\n#define TEST_DOG_PTR  testnative::dog*\n#define TEST_OWNER_PTR  testnative::owner*\n#define TEST_COUNTED_PTR  testnative::reference_counter*\n\ntestnative::dog* create_dog();\ntestnative::owner* create_owner(testnative::dog* d);\nvoid say_walk(testnative::owner* owner);\nvoid release(testnative::reference_counter* obj);\n// etc.\n```\n\nFrom the outside of the library the API is exported with opaque pointers `void*` (C structs pointers and native C99 types could be handled too).\n\n```C++\nvoid* create_dog();\nvoid* create_owner(void* d);\nvoid say_walk(void* owner);\nvoid release(void* obj);\n// etc.\n```\n\nStarting with the end in mind, from Python we want an API hiding the low level details close to the C API, in particular avoiding managing native memory via `release` C API calls, piggybacking the python GC instead.\n\n```python\ndog = Dog()\nowner = DogOwner(dog)\nowner.say_walk()\nprint(dog.position)\ndog = None # the "native dog" is still alive though, as the owner incremented the ref count\nowner = None\n```\n\nThis is doable with `refcount` and the `cffi` package. One possible design is:\n\n```python\nut_ffi = cffi.FFI()\n\nut_ffi.cdef(\'extern void* create_dog();\')\nut_ffi.cdef(\'extern void* create_owner( void* d);\')\nut_ffi.cdef(\'extern void say_walk( void* owner);\')\nut_ffi.cdef(\'extern void release( void* obj);\')\n# etc.\n\nut_dll = ut_ffi.dlopen(\'c:/path/to/test_native_library.dll\', 1) # Lazy loading\n\nclass CustomCffiNativeHandle(CffiNativeHandle):\n    def __init__(self, pointer, prior_ref_count = 0):\n        super(CustomCffiNativeHandle, self).__init__(pointer, type_id=\'\', prior_ref_count = prior_ref_count)\n\n    def _release_handle(self) -> bool:\n        ut_dll.release(self.get_handle())\n        return True\n\nclass Dog(CustomCffiNativeHandle):\n    def __init__(self, pointer = None):\n        if pointer is None:\n            pointer = ut_dll.create_dog()\n        super(Dog, self).__init__(pointer)\n    # etc.\n\nclass DogOwner(CustomCffiNativeHandle):\n\n    def __init__(self, dog):\n        super(DogOwner, self).__init__(None)\n        self._set_handle(ut_dll.create_owner(dog.get_handle()))\n        self.dog = dog\n        self.dog.add_ref() # Do note this important reference increment\n\n    def say_walk(self):\n        ut_dll.say_walk(self.get_handle())\n\n    def _release_handle(self) -> bool:\n        super(DogOwner, self)._release_handle()\n        # super(DogOwner, self)._release_handle()\n        self.dog.release()\n        return True\n```\n\n## Related work\n\n### Ancestry, acknowledgements\n\nThis python package `refcount` actually spawned from prior work for interoperability between C++, R and .NET ([R.NET](https://github.com/rdotnet/rdotnet))\n\n`refcount` features using `cffi` were also significantly informed by Kevin Plastow\'s [work](https://search.informit.com.au/documentSummary;dn=823898220073899;res=IELENG) while he was at the Australian Bureau of Meteorology; this contribution is gratefully acknowledged.\n\nIn you have native interop needs you may also want to look at:\n\n* the nuget package [dynamic-interop-dll](https://github.com/rdotnet/dynamic-interop-dll) for .NET/native interop.\n* a set of mostly c++ software [tools for interop with C/C++](https://github.com/csiro-hydroinformatics/c-interop)\n* a C# library for [generating interop glue code on top of C API glue code](https://github.com/csiro-hydroinformatics/c-api-wrapper-generation).\n\n### Other python packages\n\n`refcount` was created in part because no existing prior (Python) work could quite fit the need. There are however packages that may better address your particular need:\n\n* [infi.pyutils](https://pypi.org/project/infi.pyutils/) contains a reference counting class.\n',
    'author': 'J-M',
    'author_email': 'jean-michel.perraud@csiro.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
